@@ -18,26 +18,35 @@ from sympy import Indexed, IndexedBase
 from sympy import diff
 from sympy import log
 from sympy import preorder_traversal
+from sympy import cacheit
 
 from sympy.core.function      import AppliedUndef
 from sympy.core.function      import UndefinedFunction
 from sympy.core.compatibility import is_sequence
 
-from sympde.core.basic import CalculusFunction
-from sympde.core.basic import _coeffs_registery
-from sympde.core.basic import BasicMapping
-from sympde.core.algebra import LinearOperator
-from .space import ScalarTestFunction, VectorTestFunction, IndexedTestTrial
-from .space import ScalarField, VectorField, IndexedVectorField
+from sympde.core.basic    import CalculusFunction
+from sympde.core.basic    import _coeffs_registery
+from sympde.core.basic    import BasicMapping
+from sympde.core.algebra  import LinearOperator
 from sympde.calculus.core import minus, plus
+from sympde.calculus.core import has
+
+from .space   import ScalarTestFunction, VectorTestFunction, IndexedTestTrial
+from .space   import ScalarField, VectorField, IndexedVectorField
 
 #==============================================================================
 class DifferentialOperator(LinearOperator):
     """
     This class is a linear operator that applies the Leibniz formula
 
+    Parameters
+    ----------
+    expr : Expr
+        expr represents a Sympy expression
+
     """
     coordinate = None
+    logical    = None
 
     # this is a hack since sometimes we use Matrix in the eval of abstract
     # operators
@@ -45,11 +54,41 @@ class DifferentialOperator(LinearOperator):
         return self
 
     @classmethod
-    def eval(cls, *_args):
-        """."""
+    @cacheit
+    def eval(cls, expr):
 
-        expr = _args[0]
-        if isinstance(expr, (IndexedTestTrial, IndexedVectorField, DifferentialOperator)):
+        types = (VectorTestFunction, ScalarTestFunction,
+                DifferentialOperator, ScalarField, VectorField)
+
+        if isinstance(expr, _logical_partial_derivatives):
+            atom    = get_atom_logical_derivatives(expr)
+            indices = get_index_logical_derivatives(expr)
+
+            if cls in _logical_partial_derivatives:
+                indices[cls.coordinate] += 1
+
+            for i,n in enumerate(list(indices.values())[::-1]):
+                d = _logical_partial_derivatives[-i-1]
+                for _ in range(n):
+                    atom = d(atom, evaluate=False)
+
+            if cls in _partial_derivatives:
+                atom = cls(atom, evaluate=False)
+            elif cls not in _logical_partial_derivatives:
+                raise NotImplementedError('TODO')
+
+            return atom
+
+        if isinstance(expr, (VectorTestFunction, VectorField)):
+            n = expr.shape[0]
+            args = [cls(expr[i], evaluate=False) for i in range(0, n)]
+            args = Tuple(*args)
+            return Matrix([args])
+        elif isinstance(expr, (list, tuple, Tuple, Matrix, ImmutableDenseMatrix)):
+            args = [cls(i, evaluate=True) for i in expr]
+            args = Tuple(*args)
+            return Matrix([args])
+        elif isinstance(expr, (IndexedTestTrial, IndexedVectorField, DifferentialOperator)):
             return cls(expr, evaluate=False)
 
         elif isinstance(expr, (ScalarField, ScalarTestFunction)):
@@ -58,24 +97,29 @@ class DifferentialOperator(LinearOperator):
         elif isinstance(expr, (minus, plus)):
             return cls(expr, evaluate=False)
 
-        elif isinstance(expr, (VectorTestFunction, VectorField)):
-            n = expr.shape[0]
-            args = [cls(expr[i], evaluate=False) for i in range(0, n)]
-            args = Tuple(*args)
-            return Matrix([args])
-
         elif isinstance(expr, Indexed) and isinstance(expr.base, BasicMapping):
             return cls(expr, evaluate=False)
+        elif not has(expr, types):
+            if expr.is_number:
+                return S.Zero
 
-        elif isinstance(expr, (list, tuple, Tuple)):
-            args = [cls(i, evaluate=True) for i in expr]
-            args = Tuple(*args)
-            return Matrix([args])
+            elif isinstance(expr, Expr):
+                x = Symbol(cls.coordinate)
+                if cls.logical:
+                    M = expr.atoms(Mapping)
+                    if len(M)>0:
+                        M = list(M)[0]
+                        expr_primes = [diff(expr, M[i]) for i in range(M.rdim)]
+                        Jj = Jacobian(M)[:,cls.grad_index]
+                        expr_prime = sum([ei*Jji for ei,Jji in zip(expr_primes, Jj)])
+                        return expr_prime + diff(expr, x)
+                return diff(expr, x)
 
-        elif isinstance(expr, Add):
-            args = expr.args
-            args = [cls.eval(a) for a in expr.args]
-            return Add(*args)
+
+        if isinstance(expr, Add):
+            args = [cls(a, evaluate=True) for a in expr.args]
+            v    = Add(*args)
+            return v
 
         elif isinstance(expr, Mul):
             coeffs  = [a for a in expr.args if isinstance(a, _coeffs_registery)]
@@ -85,11 +129,11 @@ class DifferentialOperator(LinearOperator):
             if coeffs:
                 c = Mul(*coeffs)
 
-            V = S.One
+            V = S.Zero
             if vectors:
                 if len(vectors) == 1:
                     # do we need to use Mul?
-                    V = cls(Mul(vectors[0]), evaluate=True)
+                    V = cls(vectors[0], evaluate=True)
 
                 elif len(vectors) == 2:
                     a = vectors[0]
@@ -109,54 +153,38 @@ class DifferentialOperator(LinearOperator):
 
                     V = a * fb + fa * b
 
-            return Mul(c, V)
+            v = Mul(c, V)
+            return v
 
         elif isinstance(expr, Pow):
             b = expr.base
             e = expr.exp
-            return (log(b)*cls.eval(e) + e*cls.eval(b)/b) * b**e
-
-        elif isinstance(expr, Derivative):
-            x = Symbol(cls.coordinate)
-            f = expr.args[0]
-            args = list(expr.args[1:])
-            args += [x]
-            return Derivative(f, *args)
-
-        elif isinstance(expr, UndefinedFunction):
-            x = Symbol(cls.coordinate)
-            return Derivative(expr, x)
-
-        elif isinstance(expr, AppliedUndef):
-            x = Symbol(cls.coordinate)
-            return Derivative(expr, x)
-
-        elif isinstance(expr, _coeffs_registery):
-            return S.Zero
-
-        elif isinstance(expr, Expr):
-            x = Symbol(cls.coordinate)
-            return diff(expr, x)
+            v = (log(b)*cls(e, evaluate=True) + e*cls(b, evaluate=True)/b) * b**e
+            return v
 
         else:
             msg = '{expr} of type {type}'.format(expr=expr, type=type(expr))
             raise NotImplementedError(msg)
 
+    def _sympystr(self, printer):
+        sstr = printer.doprint
+        args = ','.join(sstr(i) for i in self.args)
+        return 'd{}({})'.format(sstr(self.coordinate), args)
 #==============================================================================
 class dx(DifferentialOperator):
     coordinate = 'x'
     grad_index = 0 # index in grad
-    pass
+
 
 class dy(DifferentialOperator):
     coordinate = 'y'
     grad_index = 1 # index in grad
-    pass
+
 
 class dz(DifferentialOperator):
     coordinate = 'z'
     grad_index = 2 # index in grad
-    pass
+
 
 _partial_derivatives = (dx, dy, dz)
 
@@ -164,21 +192,22 @@ _partial_derivatives = (dx, dy, dz)
 class dx1(DifferentialOperator):
     coordinate = 'x1'
     grad_index = 0 # index in grad
-    pass
+    logical    = True
 
 class dx2(DifferentialOperator):
     coordinate = 'x2'
     grad_index = 1 # index in grad
-    pass
+    logical    = True
 
 class dx3(DifferentialOperator):
     coordinate = 'x3'
     grad_index = 2 # index in grad
-    pass
+    logical    = True
 
 _logical_partial_derivatives = (dx1, dx2, dx3)
 
 #==============================================================================
+@cacheit
 def find_partial_derivatives(expr):
     """
     returns all partial derivative expressions
@@ -196,14 +225,15 @@ def find_partial_derivatives(expr):
         return args
 
     elif isinstance(expr, _partial_derivatives):
-        return [expr]
+        return (expr,)
 
     elif isinstance(expr, _logical_partial_derivatives):
-        return [expr]
+        return (expr,)
 
-    return []
+    return ()
 
 #==============================================================================
+@cacheit
 def get_number_derivatives(expr):
     """
     returns the number of partial derivatives in expr.
@@ -218,6 +248,7 @@ def get_number_derivatives(expr):
     return n
 
 #==============================================================================
+@cacheit
 def sort_partial_derivatives(expr):
     """returns the partial derivatives of an expression, sorted.
     """
@@ -260,9 +291,10 @@ def sort_partial_derivatives(expr):
         ls += d[k]
     # ...
 
-    return ls
+    return tuple(ls)
 
 #==============================================================================
+@cacheit
 def get_index_derivatives(expr):
     """
     """
@@ -288,6 +320,7 @@ def get_index_derivatives(expr):
     return d
 
 #==============================================================================
+@cacheit
 def get_atom_derivatives(expr):
     """
     """
@@ -326,6 +359,7 @@ def get_index_logical_derivatives(expr):
     return d
 
 #==============================================================================
+@cacheit
 def get_atom_logical_derivatives(expr):
     """
     """
@@ -361,7 +395,6 @@ class Dot_1d(DotBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -378,7 +411,6 @@ class Dot_2d(DotBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -407,7 +439,6 @@ class Dot_3d(DotBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -457,7 +488,6 @@ class Cross_2d(CrossBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -478,7 +508,6 @@ class Cross_3d(CrossBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -496,6 +525,7 @@ class GradBasic(CalculusFunction):
     nargs = None
     name = 'Grad'
 
+    @cacheit
     def __new__(cls, *args, **options):
         # (Try to) sympify args first
 
@@ -519,8 +549,8 @@ class GradBasic(CalculusFunction):
 class Grad_1d(GradBasic):
 
     @classmethod
+    @cacheit
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -532,8 +562,8 @@ class Grad_1d(GradBasic):
 class Grad_2d(GradBasic):
 
     @classmethod
+    @cacheit
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -557,8 +587,8 @@ class Grad_2d(GradBasic):
 class Grad_3d(GradBasic):
 
     @classmethod
+    @cacheit
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -609,7 +639,6 @@ class Curl_2d(CurlBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -622,7 +651,6 @@ class Curl_3d(CurlBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -661,7 +689,6 @@ class Rot_2d(CalculusFunction):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -700,20 +727,17 @@ class Div_1d(DivBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
 
         u = _args[0]
-
-        return dx(u)
+        return dx(u[0])
 
 class Div_2d(DivBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -726,7 +750,6 @@ class Div_3d(DivBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -765,7 +788,6 @@ class Laplace_1d(LaplaceBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -780,7 +802,6 @@ class Laplace_2d(LaplaceBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -795,7 +816,6 @@ class Laplace_3d(LaplaceBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -836,7 +856,6 @@ class Hessian_1d(HessianBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -851,7 +870,6 @@ class Hessian_2d(HessianBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -867,7 +885,6 @@ class Hessian_3d(HessianBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -904,7 +921,6 @@ class Bracket_2d(BracketBasic):
 
     @classmethod
     def eval(cls, *_args):
-        """."""
 
         if not _args:
             return
@@ -915,57 +931,285 @@ class Bracket_2d(BracketBasic):
         return dx(u)*dy(v) - dy(u)*dx(v)
 
 #==============================================================================
-# ... TODO to be removed, not used anymore
-def partial_derivative_as_symbol(expr, name=None, dim=None):
-    """Returns a Symbol from a partial derivative expression."""
-    if not isinstance(expr, _partial_derivatives):
-        raise TypeError('Expecting a partial derivative expression')
+class LogicalGrad_1d(GradBasic):
 
-    index = get_index_derivatives(expr)
-    var = get_atom_derivatives(expr)
+    @classmethod
+    def eval(cls, *_args):
 
-    if not isinstance(var, (Symbol, Indexed)):
-        raise TypeError('Expecting a Symbol, Indexed')
+        if not _args:
+            return
 
-    code = ''
-    for k,n in list(index.items()):
-        code += k*n
+        u = _args[0]
 
-    if var.is_Indexed:
-        if name is None:
-            name = var.base
+        return dx1(u)
 
-        indices = ''.join('{}'.format(i) for i in var.indices)
-        name = '{name}_{code}'.format(name=name, code=code)
-        shape = None
-        if dim:
-            shape = [dim]
-        return IndexedBase(name, shape=shape)[indices]
+class LogicalGrad_2d(GradBasic):
 
-    else:
-        if name is None:
-            name = var.name
+    @classmethod
+    def eval(cls, *_args):
 
-        name = '{name}_{code}'.format(name=name, code=code)
-        return Symbol(name)
+        if not _args:
+            return
+
+        u = _args[0]
+
+        if isinstance(u, (Tuple, Matrix, ImmutableDenseMatrix)):
+            n = len(u)
+            lines = []
+            for i in range(0, n):
+                line = [dx1(u)[0,i], dx2(u)[0,i]]
+                lines.append(line)
+            v = ImmutableDenseMatrix(lines)
+
+        else:
+            v = ImmutableDenseMatrix((dx1(u), dx2(u)))
+
+        return v
+
+class LogicalGrad_3d(GradBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+
+        if isinstance(u, (Tuple, Matrix, ImmutableDenseMatrix)):
+            n = len(u)
+            lines = []
+            for i in range(0, n):
+                line = [dx1(u)[0,i], dx2(u)[0,i], dx3(u)[0,i]]
+                lines.append(line)
+
+            v = ImmutableDenseMatrix(lines)
+
+        else:
+            v = ImmutableDenseMatrix((dx1(u), dx2(u), dx3(u)))
+
+        return v
 
 #==============================================================================
-def partial_derivative_as_str(expr):
-    """Returns a string from a partial derivative expression."""
-    if not isinstance(expr, _partial_derivatives):
-        raise TypeError('Expecting a partial derivative expression')
 
-    index = get_index_derivatives(expr)
-    var = get_atom_derivatives(expr)
+class LogicalCurl_2d(CurlBasic):
 
-    if not isinstance(var, (Symbol, Indexed)):
-        raise TypeError('Expecting a Symbol, Indexed')
+    @classmethod
+    def eval(cls, *_args):
 
-    code = ''
-    for k,n in list(index.items()):
-        code += k*n
+        if not _args:
+            return
 
-    return code
+        u = _args[0]
+
+        return dx1(u[1])-dx2(u[0])
+
+class LogicalCurl_3d(CurlBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+
+        return ImmutableDenseMatrix((dx2(u[2]) - dx3(u[1]),
+                     dx3(u[0]) - dx1(u[2]),
+                     dx1(u[1]) - dx2(u[0])))
+
+#==============================================================================
+class LogicalRot_2d(CalculusFunction):
+
+    nargs = None
+    name = 'Grad'
+
+    def __new__(cls, *args, **options):
+        # (Try to) sympify args first
+
+        if options.pop('evaluate', True):
+            r = cls.eval(*args)
+        else:
+            r = None
+
+        if r is None:
+            return Basic.__new__(cls, *args, **options)
+        else:
+            return r
+
+    def __getitem__(self, indices, **kw_args):
+        if is_sequence(indices):
+            # Special case needed because M[*my_tuple] is a syntax error.
+            return Indexed(self, *indices, **kw_args)
+        else:
+            return Indexed(self, indices, **kw_args)
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+
+        return ImmutableDenseMatrix([[dx2(u),-dx1(u)]]).T
+
+#==============================================================================
+class LogicalDiv_1d(DivBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+
+        return dx1(u[0])
+
+class LogicalDiv_2d(DivBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+
+        return dx1(u[0]) + dx2(u[1])
+
+class LogicalDiv_3d(DivBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+
+        return dx1(u[0]) + dx2(u[1]) + dx3(u[2])
+
+#==============================================================================
+class LogicalLaplace_1d(LaplaceBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+        if isinstance(u, (VectorTestFunction, VectorField)):
+            raise NotImplementedError('TODO')
+
+        return dx1(dx1(u))
+
+class LogicalLaplace_2d(LaplaceBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+        if isinstance(u, (VectorTestFunction, VectorField)):
+            raise NotImplementedError('TODO')
+
+        return dx1(dx1(u)) + dx2(dx2(u))
+
+class LogicalLaplace_3d(LaplaceBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+        if isinstance(u, (VectorTestFunction, VectorField)):
+            raise NotImplementedError('TODO')
+
+        return dx1(dx1(u)) + dx2(dx2(u)) + dx3(dx3(u))
+
+#==============================================================================
+class LogicalHessian_1d(HessianBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+        if isinstance(u, (VectorTestFunction, VectorField)):
+            raise NotImplementedError('TODO')
+
+        return dx1(dx1(u))
+
+class LogicalHessian_2d(HessianBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+        if isinstance(u, (VectorTestFunction, VectorField)):
+            raise NotImplementedError('TODO')
+
+        return ImmutableDenseMatrix([[dx1(dx1(u)), dx1(dx2(u))],
+                       [dx1(dx2(u)), dx2(dx2(u))]])
+
+class LogicalHessian_3d(HessianBasic):
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+        if isinstance(u, (VectorTestFunction, VectorField)):
+            raise NotImplementedError('TODO')
+
+        return ImmutableDenseMatrix([[dx1(dx1(u)), dx1(dx2(u)), dx1(dx3(u))],
+                       [dx1(dx2(u)), dx2(dx2(u)), dx2(dx3(u))],
+                       [dx1(dx3(u)), dx2(dx3(u)), dx3(dx3(u))]])
+
+#==============================================================================
+
+class LogicalBracket_2d(BracketBasic):
+
+    nargs = None
+    name = 'Bracket'
+
+    def __new__(cls, *args, **options):
+        # (Try to) sympify args first
+
+        if options.pop('evaluate', True):
+            r = cls.eval(*args)
+        else:
+            r = None
+
+        if r is None:
+            return Basic.__new__(cls, *args, **options)
+        else:
+            return r
+
+    @classmethod
+    def eval(cls, *_args):
+
+        if not _args:
+            return
+
+        u = _args[0]
+        v = _args[1]
+
+        return dx1(u)*dx2(v) - dx2(u)*dx1(v)
 
 #==============================================================================
 def get_index_derivatives_atom(expr, atom, verbose=False):
@@ -984,7 +1228,7 @@ def get_index_derivatives_atom(expr, atom, verbose=False):
             index = get_index_derivatives(i)
             indices.append(index)
 
-    return indices
+    return tuple(indices)
 
 #==============================================================================
 def get_index_logical_derivatives_atom(expr, atom, verbose=False):
@@ -1003,7 +1247,7 @@ def get_index_logical_derivatives_atom(expr, atom, verbose=False):
             index = get_index_logical_derivatives(i)
             indices.append(index)
 
-    return indices
+    return tuple(indices)
 
 #==============================================================================
 def get_max_partial_derivatives(expr, F=None):
@@ -1027,66 +1271,26 @@ def get_max_partial_derivatives(expr, F=None):
             if v > d[k]: d[k] = v
     return d
 
-#==============================================================================
-class LogicalGrad_1d(GradBasic):
 
-    @classmethod
-    def eval(cls, *_args):
-        """."""
+def get_max_logical_partial_derivatives(expr, F=None):
+    if F is None:
+        Fs = (list(expr.atoms(ScalarTestFunction)) +
+              list(expr.atoms(VectorTestFunction)) +
+              list(expr.atoms(IndexedTestTrial)) +
+              list(expr.atoms(VectorField)) +
+              list(expr.atoms(IndexedVectorField)) +
+              list(expr.atoms(ScalarField)))
 
-        if not _args:
-            return
+        indices = []
+        for F in Fs:
+            indices += get_index_logical_derivatives_atom(expr, F)
+    else:
+        indices = get_index_logical_derivatives_atom(expr, F)
 
-        u = _args[0]
+    d = {'x1':0, 'x2':0, 'x3':0}
+    for dd in indices:
+        for k,v in dd.items():
+            if v > d[k]: d[k] = v
+    return d
 
-        return dx1(u)
-
-class LogicalGrad_2d(GradBasic):
-
-    @classmethod
-    def eval(cls, *_args):
-        """."""
-
-        if not _args:
-            return
-
-        u = _args[0]
-
-        if isinstance(u, Tuple):
-            n = len(u)
-            lines = []
-            for i in range(0, n):
-                line = [dx1(u)[0,i], dx2(u)[0,i]]
-                lines.append(line)
-
-            v = Matrix(lines)
-
-        else:
-            v = Tuple(dx1(u), dx2(u))
-
-        return v
-
-class LogicalGrad_3d(GradBasic):
-
-    @classmethod
-    def eval(cls, *_args):
-        """."""
-
-        if not _args:
-            return
-
-        u = _args[0]
-
-        if isinstance(u, Tuple):
-            n = len(u)
-            lines = []
-            for i in range(0, n):
-                line = [dx1(u)[0,i], dx2(u)[0,i], dx3(u)[0,i]]
-                lines.append(line)
-
-            v = Matrix(lines)
-
-        else:
-            v = Tuple(dx1(u), dx2(u), dx3(u))
-
-        return v
+from .mapping import Mapping, Jacobian
