@@ -19,11 +19,15 @@ from sympy.core import Add, Mul, Pow
 from sympy.core.expr import AtomicExpr
 
 from sympde.core.basic import CalculusFunction
-from .basic import BasicDomain, InteriorDomain, Boundary, Union, Connectivity
-from .basic import Interval
-from .basic import ProductDomain
+from .basic            import BasicDomain, InteriorDomain, Boundary, Union, Connectivity
+from .basic            import Interval, Interface
+from .basic            import ProductDomain
+
+# TODO fix circular dependency between domain and mapping
 
 # TODO add pdim
+
+iterable_types = (tuple, list, Tuple, Union)
 #==============================================================================
 class Domain(BasicDomain):
     """
@@ -35,8 +39,8 @@ class Domain(BasicDomain):
 
     """
 
-    def __new__(cls, name, interiors=None, boundaries=None, dim=None,
-                connectivity=None):
+    def __new__(cls, name, *, interiors=None, boundaries=None, dim=None,
+                connectivity=None, mapping=None, logical_domain=None):
         # ...
         if not isinstance(name, str):
             raise TypeError('> name must be a string')
@@ -49,8 +53,7 @@ class Domain(BasicDomain):
 
         # ...
         if not( interiors is None ):
-            if ( not isinstance( interiors, (tuple, list, Tuple)) and
-                 not isinstance( interiors, InteriorDomain) ):
+            if not isinstance( interiors, (*iterable_types, InteriorDomain)):
                 raise TypeError('> Expecting an iterable or a InteriorDomain')
 
             if isinstance( interiors, InteriorDomain ):
@@ -59,8 +62,8 @@ class Domain(BasicDomain):
             else:
                 new_interiors = []
                 for i in interiors:
-                    if isinstance(i , Union):
-                        new_interiors += list(i.as_tuple())
+                    if isinstance(i , iterable_types):
+                        new_interiors += list(i)
                     else:
                         new_interiors.append(i)
 
@@ -74,8 +77,7 @@ class Domain(BasicDomain):
 
         # ...
         if not( boundaries is None ):
-            if ( not isinstance( boundaries, (tuple, list, Tuple)) and
-                 not isinstance( boundaries, Boundary) ):
+            if not isinstance( boundaries, (*iterable_types, Boundary)):
                 raise TypeError('> Expecting an iterable or a Boundary')
 
             if isinstance( boundaries, Boundary ):
@@ -107,26 +109,22 @@ class Domain(BasicDomain):
         if len(interiors) == 0 and dim is None:
             raise TypeError('No interior domain found')
 
-        elif len(interiors) == 1:
-            interiors = interiors[0]
-            dtype = interiors.dtype
-            dim   = interiors.dim
-
-        elif len(interiors) > 1:
+        else:
             dtype = interiors[0].dtype
             dim   = interiors[0].dim
             interiors = Union(*interiors)
 
-        # ...
-        if len(boundaries) > 1:
-            boundaries = Union(*boundaries)
-        # ...
+        assert mapping is None and logical_domain is None or \
+        mapping is not None and logical_domain  is not None
 
-        obj = Basic.__new__(cls, name, interiors, boundaries)
-        obj._connectivity = connectivity
-        obj._dtype        = dtype
-        obj._dim          = dim
+        # ...
+        boundaries = Union(*boundaries)
 
+        obj = Basic.__new__(cls, name, interiors, boundaries, mapping)
+        obj._connectivity   = connectivity
+        obj._dtype          = dtype
+        obj._dim            = dim
+        obj._logical_domain = logical_domain
         return obj
 
     @property
@@ -141,6 +139,14 @@ class Domain(BasicDomain):
     def boundary(self):
         return self.args[2]
 
+    @property
+    def mapping(self):
+        return self.args[3]
+
+    @property
+    def logical_domain(self):
+        return self._logical_domain
+
     @property
     def connectivity(self):
         return self._connectivity
@@ -175,7 +181,10 @@ class Domain(BasicDomain):
 
     def _sympystr(self, printer):
         sstr = printer.doprint
-        return '{}'.format(sstr(self.name))
+        if self.mapping:
+            return '{}({})'.format(sstr(self.mapping), sstr(self.name))
+        else:
+            return '{}'.format(sstr(self.name))
 
     def get_boundary(self, name=None, axis=None, ext=None):
         """return boundary by name or (axis, ext)."""
@@ -284,6 +293,7 @@ class Domain(BasicDomain):
         if not(ext == '.h5'):
             raise ValueError('> Only h5 files are supported')
         # ...
+        from sympde.topology.mapping import Mapping
 
         h5  = h5py.File( filename, mode='r' )
         yml = yaml.load( h5['topology.yml'][()], Loader=yaml.SafeLoader )
@@ -294,12 +304,13 @@ class Domain(BasicDomain):
         d_interior     = yml['interior']
         d_boundary     = yml['boundary']
         d_connectivity = yml['connectivity']
+        mapping        = Mapping('{}_mapping'.format(domain_name), int(dim))
 
         if dtype == 'None': dtype = None
 
         if dtype is not None:
             constructor = globals()[dtype['type']]
-            return constructor(domain_name, **dtype['parameters'])
+            return mapping(constructor(domain_name, **dtype['parameters']))
 
         # ... create sympde InteriorDomain (s)
         interior = [InteriorDomain(i['name'], dim=dim) for i in d_interior]
@@ -341,30 +352,42 @@ class Domain(BasicDomain):
                 bnd   = Boundary(name, patch)
                 bnds.append(bnd)
 
-            connectivity[edge] = bnds
+            connectivity[edge] = Interface(edge, bnds[0], bnds[1])
         # ...
 
-        return Domain.__new__(cls, domain_name,
+        obj = Domain.__new__(cls, domain_name,
                               interiors=interior,
                               boundaries=boundary,
                               connectivity=connectivity)
+        return mapping(obj)
 
     def join(self, other, name, bnd_minus=None, bnd_plus=None):
-        # ... interiors
-        interiors = [self.interior, other.interior]
-        # ...
 
+        from sympde.topology.mapping import InterfaceMapping
         # ... connectivity
         connectivity = Connectivity()
         # TODO be careful with '|' in psydac
         if bnd_minus and bnd_plus:
-            connectivity['{l}|{r}'.format(l=bnd_minus.domain.name, r=bnd_plus.domain.name)] = (bnd_minus, bnd_plus)
+
+            if bnd_minus.mapping and bnd_plus.mapping:
+                int_map            = InterfaceMapping(bnd_minus.mapping , bnd_plus.mapping)
+                a,b                = bnd_minus.logical_domain, bnd_plus.logical_domain
+                l_name             = '{l}|{r}'.format(l=a.domain.name, r=b.domain.name)
+                int_logical_domain = Interface(l_name, a,b)
+            else:
+                int_map            = None
+                int_logical_domain = None
+
+            name               = '{l}|{r}'.format(l=bnd_minus.domain.name, r=bnd_plus.domain.name)
+            connectivity[name] = Interface(name, bnd_minus, bnd_plus,
+                                           mapping=int_map,
+                                           logical_domain=int_logical_domain)
+
         for k,v in self.connectivity.items():
             connectivity[k] = v
 
         for k,v in other.connectivity.items():
             connectivity[k] = v
-        # ...
 
         # ... boundary
         boundaries_minus = self.boundary.complement(bnd_minus)
@@ -372,11 +395,32 @@ class Domain(BasicDomain):
 
         boundaries = Union(boundaries_minus, boundaries_plus)
         boundaries = boundaries.as_tuple()
+
+        # ... interiors
+        interiors       = Union(self.interior, other.interior)
+        if all(e.mapping for e in interiors):
+            logical_interiors    = Union(*[e.logical_domain for e in interiors])
+            logical_boundaries   = [e.logical_domain for e in boundaries]
+            logical_connectivity = Connectivity()
+            for k,v in connectivity.items():
+                logical_connectivity[name] = v.logical_domain
+
+            mapping        = tuple(e.mapping for e in interiors)
+            logical_domain = Domain(name,
+                            interiors=logical_interiors,
+                            boundaries=logical_boundaries,
+                            connectivity=logical_connectivity)
+        else:
+            mapping              = None
+            logical_domain       = None
+
         # ...
         return Domain(name,
                       interiors=interiors,
                       boundaries=boundaries,
-                      connectivity=connectivity)
+                      connectivity=connectivity,
+                      mapping=mapping,
+                      logical_domain=logical_domain)
 
 #==============================================================================
 class PeriodicDomain(BasicDomain):
@@ -421,13 +465,21 @@ class PeriodicDomain(BasicDomain):
         return self.domain.coordinates
 
     def __hash__(self):
-        return self.domain.__hash__() + self._periods.__hash__()
+        return hash((self._domain, self._periods))
 
 
 #==============================================================================
 class NCubeInterior(InteriorDomain):
-    _min_coords = None
-    _max_coords = None
+
+    def __new__(cls, name, dim=None, dtype=None, min_coords=None, max_coords=None,
+                    mapping=None, logical_domain=None):
+
+        obj = InteriorDomain.__new__(cls, name, dim=dim, dtype=dtype,
+                    mapping=mapping, logical_domain=logical_domain)
+        obj._min_coords = min_coords
+        obj._max_coords = max_coords
+        return obj
+
     @property
     def min_coords(self):
         return self._min_coords
@@ -447,8 +499,8 @@ class NCube(Domain):
 
         assert isinstance(name, str)
         assert isinstance(dim, (int, Integer))
-        assert isinstance(min_coords, (tuple, list, Tuple))
-        assert isinstance(max_coords, (tuple, list, Tuple))
+        assert isinstance(min_coords, iterable_types[:-1])
+        assert isinstance(max_coords, iterable_types[:-1])
 
         if not name:
             raise ValueError("Name must be provided")
@@ -462,11 +514,7 @@ class NCube(Domain):
         if not all(xmin < xmax for xmin, xmax in zip(min_coords, max_coords)):
             raise ValueError("Min coordinates must be smaller than max")
 
-        # Choose coordinate names. TODO: use unique convention
-        if dim <= 3:
-            coord_names = ('x', 'y', 'z')[:dim]
-        else:
-            coord_names = 'x1:{}'.format(dim + 1)
+        coord_names = 'x1:{}'.format(dim + 1)
 
         coordinates = symbols(coord_names)
         intervals   = [Interval('{}_{}'.format(name, c.name), coordinate=c, bounds=(xmin, xmax))
@@ -510,9 +558,7 @@ class NCube(Domain):
         else:
             interior = ProductDomain(*intervals, name=name)
 
-        interior             = NCubeInterior(interior, dtype=dtype)
-        interior._min_coords = tuple(min_coords)
-        interior._max_coords = tuple(max_coords)
+        interior = NCubeInterior(interior, dtype=dtype, min_coords=tuple(min_coords), max_coords=tuple(max_coords))
 
         boundaries = []
         i = 1
@@ -598,7 +644,7 @@ class Cube(NCube):
 
 #==============================================================================
 class BoundaryVector(IndexedBase):
-    pass
+    is_commutative = False
 
 class NormalVector(BoundaryVector):
     pass
@@ -738,3 +784,4 @@ def split(domain, value):
     else:
         raise NotImplementedError('TODO')
 
+
